'use strict';
var SocketP = (function() {//
    var Socket = new Class({
	initialize: function(url, emit_t, Debug) {
	    this.url = url;
	    this.emitter = emit_t;
	    this.socket = undefined;
	    this.waits = {};
	    this.debug = Debug;
	},
	// create socket with promise
	connect: function() {
	    var self = this;
	    if (this.socket) {
		return Promise.resolve({ws:self});
	    }
	    return new Promise(function(resolve, reject) {
		var finished = false;
		function finish(arg) {
		    finished = true;
		    return arg;
		}
		try {
		    self.socket = new WebSocket(self.url);
		}
		catch (e) {
		    self.socket = undefined;
		    return finish(reject(e));
		}
		if (!self.socket) {
		    self.socket = undefined;
		    return finish(reject(Error("Could not open websocket")));
		}
		self.socket.onmessage = function(event) {
		    var msg;
		    try {
			if (self.debug) { self.debug("<<", event.data); }
			msg = JSON.parse(event.data);
		    }
		    catch (e) {
			console.log("no valid json");
			return;
		    }
		    if (msg.id !== undefined && self.waits[msg.id] !== undefined) {
			return (self.waits[msg.id])({ws:self, msg:msg});
		    }
		    else {
			var k;
			for (k in msg) {
			    self.emitter(k, msg[k]);
			}
		    }
		};
		self.socket.onerror = function(event) {
		    if (!finished) {
			return finish(reject(Error(event.message || event.type)));
		    }
		    self.emitter("ws error", event.message || event.type);
		    self.socket = undefined;
		};
		self.socket.onclose = function(event) {
		    if (!finished) {
			return finish(reject(Error(event.reason || event.type)));
		    }
		    self.waits = {};
		    self.emitter("ws closed", event.reason || event.type);
		    self.socket = undefined;
		};
		self.socket.onopen = function(event) {
		    return finish(resolve({ws:self}));
		};
	    });
	},
	// send raw message
	send: function(message) {
	    if (this.debug) { this.debug(">>", message); }
	    return this.socket.send(message);
	},
	// send Json message as promise
	sendj: function(message) {
	    var self = this;
	    return new Promise(function(resolve) {
		self.send(JSON.stringify(message));
		if (message.id !== undefined) {
		    self.waits[message.id] = resolve;
		}
		else {
		    resolve(true);
		}
	    });
	},
	// send json message with automatic ID as Promise
	sendp: function(message) {
	    do {
		message.id = Math.random().toString(36).substr(2);
	    } while (this.waits[message.id] !== undefined);
	    return this.sendj(message);
	}
    });

    return Socket;
})();
